/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use std::collections::{BTreeMap, BTreeSet};
use std::fmt::Write;

pub fn bytes_to_hexstring(chunkid: &[u8]) -> String {
    let mut s = String::with_capacity(chunkid.len() * 2);
    for &byte in chunkid.iter() {
        write!(&mut s, "{:02x}", byte).expect("Unable to write");
    }
    s
}

pub fn get_number_of_workers(first_try_env_name: &str) -> u8 {
    let from_env_var = std::env::var(first_try_env_name)
        .ok()
        .or_else(|| std::env::var("YAMA_WORKERS").ok());
    if let Some(from_env_var) = from_env_var {
        from_env_var
            .parse()
            .expect("Number of workers should be an integer from 1 to 255.")
    } else {
        let number = num_cpus::get();
        if number > u8::MAX.into() {
            u8::MAX
        } else {
            number as u8
        }
    }
}

#[derive(Clone, Debug)]
pub struct LruMap<K, V> {
    capacity: usize,
    last_access: BTreeSet<(u64, K)>,
    items: BTreeMap<K, (V, u64)>,
    counter: u64,
}

impl<K: Ord + Clone, V> LruMap<K, V> {
    pub fn new(capacity: usize) -> LruMap<K, V> {
        LruMap {
            capacity,
            last_access: BTreeSet::new(),
            items: BTreeMap::new(),
            counter: 0,
        }
    }

    /// Gets an item from the LRU map.
    pub fn get(&mut self, key: &K) -> Option<&V> {
        match self.items.get_mut(key) {
            Some((value, last_used_instant)) => {
                assert!(
                    self.last_access.remove(&(*last_used_instant, key.clone())),
                    "Corrupt LRU map: freshen not correct."
                );
                let new_instant = self.counter;
                self.counter += 1;
                self.last_access.insert((new_instant, key.clone()));
                *last_used_instant = new_instant;
                Some(value)
            }
            None => None,
        }
    }

    pub fn insert(&mut self, key: K, value: V) -> Option<V> {
        let new_instant = self.counter;
        self.counter += 1;

        let retval = match self.items.insert(key.clone(), (value, new_instant)) {
            Some((old_entry, old_instant)) => {
                assert!(
                    self.last_access.remove(&(old_instant, key.clone())),
                    "Corrupt LRU map: insert not correct."
                );
                Some(old_entry)
            }
            None => None,
        };
        self.last_access.insert((new_instant, key));

        if retval.is_none() {
            // We didn't replace any item, so we have grown by 1.
            // Check if we need to evict.
            if self.items.len() > self.capacity {
                self.evict();
            }
        }

        retval
    }

    pub fn evict(&mut self) -> Option<(K, V)> {
        if let Some(first_entry) = self.last_access.iter().next().cloned() {
            self.last_access.remove(&first_entry);
            let (_, key) = first_entry;
            let (value, _) = self
                .items
                .remove(&key)
                .expect("Corrupt LRU map: last access and items out of sync");

            Some((key, value))
        } else {
            None
        }
    }
}

#[cfg(test)]
mod test {
    use crate::utils::LruMap;

    #[test]
    fn test_lru_map() {
        let mut lmap = LruMap::new(3);
        lmap.insert(1, 1);
        lmap.insert(2, 1);
        lmap.insert(3, 1);
        assert_eq!(lmap.get(&1), Some(&1));
        lmap.insert(4, 1);
        assert_eq!(lmap.get(&2), None);
    }
}
